import cupy as cp

def ft2d(image):
    ft = cp.fft.fft2(image)
    return cp.fft.fftshift(ft)
def ift2d(image):
    ift = cp.fft.ifftshift(image)
    ift = cp.fft.ifft2(ift)
    return ift.real

def normalize(image):
    return (image - cp.min(image)) / (cp.max(image) - cp.min(image))

def low_pass_filter_gaussian(image, cutoff_angstrom, pixel_size):
    '''
    low pass filter for the image using Gaussian filter.
    Args:
        image: image
        cutoff_angstrom: cutoff angstrom
        pixel_size: pixel size (angstrom)
    Returns:
        ift: low pass filtered image
    '''
    rows, cols = image.shape
    freq_x = cp.linspace(-0.5, 0.5, cols)
    freq_y = cp.linspace(-0.5, 0.5, rows)
    fx, fy = cp.meshgrid(freq_x, freq_y)
    freq_radius = cp.sqrt(fx**2 + fy**2)
    
    cutoff_frequency = 2 * pixel_size / cutoff_angstrom
    mask = cp.exp(-(freq_radius**2) / (2 * (cutoff_frequency)**2))

    image = normalize(image)

    f_shifted = ft2d(image)
    f_filtered_shifted = f_shifted * mask
    ift = ift2d(f_filtered_shifted)

    ift = normalize(ift)
    return ift.get()


import os

def get_lib_path(lib_name):
    '''
    Get the path to the library.
    Args:
        lib_name: name of the library
    Returns:
        path: path to the library
    '''
    return os.path.join(os.path.dirname(__file__), 'lib', f'{lib_name}.so')


from .carbon_film_detector import *
from .detector_for_dynamo import multi_mrc_processing_dynamo
from .detector_for_cryosparc import multi_mrc_processing_cryosparc
import argparse
import os
import sys
import setproctitle
    
def main():
    # set process name
    setproctitle.setproctitle('cryoCOFI')
    parser = argparse.ArgumentParser(description='''
    -----------------------------------
    cryoCOFI: CarbOn FIlm detector for cryo-EM images
    -----------------------------------
    ''',
    formatter_class=argparse.RawTextHelpFormatter,
    epilog='''
    -----------------------------------
    Email: zhen.victor.huang@gmail.com if you have any questions. 
    Please visit https://github.com/ZhenHuangLab/cryoCOFI for more information.
    -----------------------------------
    ''')
    subparsers = parser.add_subparsers(dest='command', help='Please specify the command to run!')
    readmrc_parser = subparsers.add_parser('readmrc', help='Read MRC file and detect carbon film')
    readdynamo_parser = subparsers.add_parser('readdynamo', help='Read Dynamo doc and tbl file and output a new tbl file without particles inside the carbon film')
    readcs_parser = subparsers.add_parser('readcs', help='Read CryoSPARC cs file and output the new cs file')

    # Add arguments to subparsers
    readmrc_args(readmrc_parser)
    readcs_args(readcs_parser)
    readdynamo_args(readdynamo_parser)

    try:
        args = parser.parse_args()
        
        if args.command == 'readmrc':
            handle_readmrc(args)
        elif args.command == 'readcs':
            handle_readcs(args)
        elif args.command == 'readdynamo':
            handle_readdynamo(args)
        else:
            parser.print_help()
            
    except (argparse.ArgumentError, argparse.ArgumentTypeError) as e:
        print(f"\nError: {str(e)}\n")
        if args.command == 'readmrc':
            readmrc_parser.print_help()
        elif args.command == 'readcs':
            readcs_parser.print_help()
        elif args.command == 'readdynamo':
            readdynamo_parser.print_help()
        else:
            parser.print_help()
        sys.exit(1)

def readmrc_args(parser):
    parser.add_argument('--input', '-i', type=str, required=True, help='Input MRC file')
    parser.add_argument('--lowpass', '-lp', type=int, default=200, help='Low pass filter cutoff angstrom')
    parser.add_argument('--detector_type', '-dt', type=str, default='bicanny', help='''Specify the detector type: bicanny or canny. Default is bicanny. 
                                    For tomograms, bicanny is recommended; for SPA, canny is recommended.''')
    parser.add_argument('--kernel_radius', '-kr', type=int, default=5, help='Kernel radius for bilateral filter')
    parser.add_argument('--sigma_color', '-sc', type=float, default=10.0, help='Sigma color for bilateral filter')
    parser.add_argument('--sigma_space', '-ss', type=float, default=10.0, help='Sigma space for bilateral filter')
    parser.add_argument('--canny_kernel', '-ck', type=int, default=2, help='Canny kernel size for edge detection')
    parser.add_argument('--diameter', '-d', type=int, default=12000, help='Carbon Hole Diameter in Angstrom')
    parser.add_argument('--map_cropping', '-mc', type=int, default=20, help='Removing edge pixels and cropping the image')
    parser.add_argument('--dist_thr_inside_edge', '-dte', type=int, default=20, help='Distance threshold for inside edge pixels')
    parser.add_argument('--mode_threshold', '-mt', type=float, default=0, help='Mode threshold for finding the carbon film edge')
    parser.add_argument('--edge_quotient_threshold', '-eqt', type=float, required=True, help='Edge quotient threshold for finding the carbon film edge')
    parser.add_argument('--show_fig', '-sf', action='store_true', default=False, help='Show figures if specified')
    parser.add_argument('--verbose', '-v', action='store_true', default=False, help='Show verbose information if specified')
    parser.add_argument('--gpu', type=int, default=0, help='GPU device number to use. Default is 0 and start from 0.')

def readcs_args(parser):
    parser.add_argument('--cs_path', '-i', type=str, required=True, help='Input CryoSPARC .cs file')
    parser.add_argument('--out_path', '-o', type=str, required=True, help='Output CryoSPARC .cs file')
    parser.add_argument('--low_pass', '-lp', type=int, default=300, help='Low pass filter cutoff angstrom')
    parser.add_argument('--detector_type', '-dt', type=str, default='canny', help='''Specify the detector type: bicanny or canny. Default is bicanny. 
                                For tomograms, bicanny is recommended; for SPA, canny is recommended.''')
    parser.add_argument('--kernel_radius', '-kr', type=int, default=5, help='Kernel radius for bilateral filter')
    parser.add_argument('--sigma_color', '-sc', type=float, default=10.0, help='Sigma color for bilateral filter')
    parser.add_argument('--sigma_space', '-ss', type=float, default=10.0, help='Sigma space for bilateral filter')
    parser.add_argument('--canny_kernel', '-ck', type=int, default=2, help='Canny kernel size for edge detection')
    parser.add_argument('--diameter', '-d', type=int, default=12000, help='Carbon Hole Diameter in Angstrom')
    parser.add_argument('--map_cropping', '-mc', type=int, default=20, help='Removing edge pixels and cropping the image')
    parser.add_argument('--dist_thr_inside_edge', '-dte', type=int, default=20, help='Distance threshold for inside edge pixels')
    parser.add_argument('--mode_threshold', '-mt', type=float, default=0, help='Mode threshold for finding the carbon film edge')
    parser.add_argument('--edge_quotient_threshold', '-eqt', type=float, required=True, help='Edge quotient threshold for finding the carbon film edge. Please specify it dataset by dataset.')
    parser.add_argument('--verbose', '-v', action='store_true', default=False, help='Show verbose information if specified')
    parser.add_argument('--gpu', type=int, default=0, help='GPU device number to use. Default is 0 and start from 0.')

def readdynamo_args(parser):
    parser.add_argument('--doc_path', '-doc', type=str, required=True, help='Input Dynamo .doc file')
    parser.add_argument('--tbl_path', '-tbl', type=str, required=True, help='Input Dynamo .tbl file')
    parser.add_argument('--out_path', '-o', type=str, required=True, help='Output Dynamo .tbl file')
    parser.add_argument('--low_pass', '-lp', type=int, default=200, help='Low pass filter cutoff angstrom')
    parser.add_argument('--detector_type', '-dt', type=str, default='bicanny', help='''Specify the detector type: bicanny or canny. Default is bicanny. 
                                For tomograms, bicanny is recommended; for SPA, canny is recommended.''')
    parser.add_argument('--kernel_radius', '-kr', type=int, default=5, help='Kernel radius for bilateral filter')
    parser.add_argument('--sigma_color', '-sc', type=float, default=10.0, help='Sigma color for bilateral filter')
    parser.add_argument('--sigma_space', '-ss', type=float, default=10.0, help='Sigma space for bilateral filter')
    parser.add_argument('--canny_kernel', '-ck', type=int, default=2, help='Canny kernel size for edge detection')
    parser.add_argument('--diameter', '-d', type=int, default=12000, help='Carbon Hole Diameter in Angstrom')
    parser.add_argument('--map_cropping', '-mc', type=int, default=20, help='Removing edge pixels and cropping the image')
    parser.add_argument('--dist_thr_inside_edge', '-dte', type=int, default=20, help='Distance threshold for inside edge pixels')
    parser.add_argument('--mode_threshold', '-mt', type=float, default=0, help='Mode threshold for finding the carbon film edge')
    parser.add_argument('--edge_quotient_threshold', '-eqt', type=float, default=6, help='Edge quotient threshold for finding the carbon film edge')
    parser.add_argument('--verbose', '-v', action='store_true', default=False, help='Show verbose information if specified')
    parser.add_argument('--gpu', type=int, default=0, help='GPU device number to use. Default is 0 and start from 0.')

def handle_readmrc(args):
    os.environ['CUDA_VISIBLE_DEVICES'] = str(args.gpu)
    detector_for_mrc(
        args.input,
        args.lowpass,
        args.detector_type,
        args.kernel_radius,
        args.sigma_color,
        args.sigma_space,
        args.canny_kernel,
        args.diameter,
        args.map_cropping,
        args.dist_thr_inside_edge,
        args.mode_threshold,
        args.edge_quotient_threshold,
        args.show_fig,
        args.verbose
    )

def handle_readcs(args):
    os.environ['CUDA_VISIBLE_DEVICES'] = str(args.gpu)
    multi_mrc_processing_cryosparc(
        args.cs_path,
        args.out_path,
        args.low_pass,
        args.detector_type,
        args.kernel_radius,
        args.sigma_color,
        args.sigma_space,
        args.canny_kernel,
        args.diameter,
        args.map_cropping,
        args.dist_thr_inside_edge,
        args.mode_threshold,
        args.edge_quotient_threshold,
        args.verbose
    )

def handle_readdynamo(args):
    os.environ['CUDA_VISIBLE_DEVICES'] = str(args.gpu)
    multi_mrc_processing_dynamo(
        args.doc_path,
        args.tbl_path,
        args.out_path,
        args.low_pass,
        args.detector_type,
        args.kernel_radius,
        args.sigma_color,
        args.sigma_space,
        args.canny_kernel,
        args.diameter,
        args.map_cropping,
        args.dist_thr_inside_edge,
        args.mode_threshold,
        args.edge_quotient_threshold,
        args.verbose
    )

if __name__ == '__main__':
    main()



from .carbon_film_detector import *
import tqdm
import numpy as np
import os
from cryosparc.dataset import Dataset


def multi_mrc_processing_cryosparc(
        cs_path,
        out_path,
        low_pass,
        detector_type,
        kernel_radius,
        sigma_color,
        sigma_space,
        canny_kernel,
        diameter,
        map_cropping,
        dist_thr_inside_edge,
        mode_threshold,
        edge_quotient_threshold,
        verbose):
    '''
    Process the mrc files from the CryoSPARC cs file.
    Args:
        cs_path: path to the CryoSPARC cs file
        out_path: path to the output file (tbl file)
        low_pass: low pass filter (angstrom)
        kernel_radius: kernel radius (pixels)
        sigma_color: sigma color
        sigma_space: sigma space
        canny_kernel: canny detector kernel size
        diameter: diameter (angstrom)
        map_cropping: map cropping (pixels)
        dist_thr_inside_edge: distance inside edge (pixels)
        mode_threshold: mode threshold
        edge_quotient_threshold: edge quotient threshold
        verbose: whether to show the print information
    '''
    particle_dset = Dataset.load(cs_path)
    # Initialize a new empty Dataset
    particle_dset_modified = Dataset.append_many()
    particle_dset_modified = Dataset.innerjoin(particle_dset, particle_dset_modified)
    
    micrograph_paths = np.array(particle_dset['location/micrograph_path'])
    # remove the duplicate micrograph paths
    micrograph_paths = np.unique(micrograph_paths)
    
    # inner progress bar
    inner_pbar = tqdm.tqdm([], desc="Screening particles", 
                          position=1, dynamic_ncols=True, unit="ptcl", leave=True)

    try:
        for micrograph_path in tqdm.tqdm(micrograph_paths, desc="Processing micrographs", 
                                        position=0, dynamic_ncols=True, unit="mg",
                                        leave=True):
            particle_dset_slice = particle_dset.mask(np.array(particle_dset['location/micrograph_path']) == micrograph_path)
            
            if not os.path.exists(micrograph_path):
                print(f"Warning: Skip micrograph {micrograph_path} because not exist.")
                particle_dset_modified.extend(particle_dset_slice)
                continue

            mg_mask = detector_for_mrc(micrograph_path,
                                low_pass,
                                detector_type,
                                kernel_radius,
                                sigma_color,
                                sigma_space,
                                canny_kernel,
                                diameter,
                                map_cropping,
                                dist_thr_inside_edge,
                                mode_threshold,
                                edge_quotient_threshold,
                                show_fig=False,
                                verbose=verbose)

            if mg_mask is False:
                particle_dset_modified.extend(particle_dset_slice)
                inner_pbar.reset(total=1)
                inner_pbar.bar_format = "{desc}: NA%|{bar:5} no carbon film detected {bar:5}|NA/NA [{elapsed}<NA, NA ptcls]"
                inner_pbar.update(0)
                if verbose:
                    print(f"Skip micrograph {micrograph_path} because no carbon film detected.")
            
            else:
                # if carbon film detected, screening the particles
                if verbose:
                    print(f"Processing micrograph {micrograph_path} with carbon film detected.")
                inner_pbar.bar_format = None
                inner_pbar.reset(total=len(particle_dset_slice))
                # screening the particles
                for particle_index in range(len(particle_dset_slice)):
                    inner_pbar.update(1)
                    micrograph_ny = particle_dset_slice['location/micrograph_shape'][particle_index][0]
                    micrograph_nx = particle_dset_slice['location/micrograph_shape'][particle_index][1]
                    location_x = particle_dset_slice['location/center_x_frac'][particle_index] * micrograph_nx
                    location_y = particle_dset_slice['location/center_y_frac'][particle_index] * micrograph_ny
                    if mg_mask[int(location_y), int(location_x)] == 1:
                        particle_dset_modified.extend(particle_dset_slice.slice(start=particle_index, stop=particle_index+1))
    
    except KeyboardInterrupt:
        inner_pbar.bar_format = None
        inner_pbar.close()
        raise
    finally:
        inner_pbar.close()
    
    # save the modified cs file
    particle_dset_modified.save(out_path)
    print(f"New cs file saved to {out_path}.")
                                


import numpy as np
import ctypes
from cryoCOFI import get_lib_path

def find_highest_density(img, mask, inside_flags=(1, 4), outside_flags=(0,)):
    '''
    Find the density of the highest frequency inside the arc.
    Args:
        img: image
        mask: mask
        inside_flags: flags for the inside
        outside_flags: flags for the outside
    Returns:
        diff: difference
    '''

    if img.ndim != 2 or mask.ndim != 2:
        raise ValueError("Both img and mask must be 2D numpy arrays.")
    
    # Load the CUDA library
    cuda_lib = ctypes.CDLL(get_lib_path('find_highest_density_double'))
    
    # Define the function prototype
    cuda_lib.find_highest_density_cuda.argtypes = [
        np.ctypeslib.ndpointer(dtype=np.float64, flags="C_CONTIGUOUS"),
        np.ctypeslib.ndpointer(dtype=np.int32, flags="C_CONTIGUOUS"),
        ctypes.c_int,
        np.ctypeslib.ndpointer(dtype=np.int32, flags="C_CONTIGUOUS"),
        ctypes.c_int,
        np.ctypeslib.ndpointer(dtype=np.int32, flags="C_CONTIGUOUS"),
        ctypes.c_int
    ]
    cuda_lib.find_highest_density_cuda.restype = ctypes.c_double

    # Ensure the arrays are contiguous and have the correct data types
    img_flat = np.ascontiguousarray(img.flatten(), dtype=np.float64)
    mask_flat = np.ascontiguousarray(mask.flatten(), dtype=np.int32)
    
    # Convert inside_flags and outside_flags to numpy arrays
    inside_flags = np.atleast_1d(inside_flags).astype(np.int32)
    outside_flags = np.atleast_1d(outside_flags).astype(np.int32)
    
    # Call the CUDA function
    diff = cuda_lib.find_highest_density_cuda(
        img_flat, mask_flat, img_flat.size,
        inside_flags, len(inside_flags),
        outside_flags, len(outside_flags)
    )
    
    return diff



import cupy as cp
import mrcfile

def read_pixel_size(mrc_path):
    '''
    Read the pixel size from the mrc file.
    Args:
        mrc_path: path to the mrc file
    Returns:
        pixel_size: pixel size
    '''
    with mrcfile.open(mrc_path) as mrc:
        pixel_size = mrc.voxel_size.x
    return pixel_size

def flag_masking(img, a, b, radius, map_cropping=20, dist_thr_inside_edge=20, map_cropping_flag=2, edge_flag=3, flag_for_dist_thr_inside_edge=4):
    '''
    Generate a mask for calculating the average density inside and outside the circle.
    Args:
        img: image
        a: center x
        b: center y
        radius: radius
        map_cropping: size of the map cropping (pixels)
        dist_thr_inside_edge: distance threshold inside edge (pixels). 
                              Extending the edge so that particles within this distance will also be removed.
        map_cropping_flag: flag for the map cropping
        edge_flag: flag for the edge
        flag_for_dist_thr_inside_edge: flag for the distance threshold inside edge
    Returns:
        mask: mask
    '''
    mask = cp.zeros_like(img)
    y, x = cp.ogrid[:img.shape[0], :img.shape[1]]
    distance_map = cp.sqrt((x - b)**2 + (y - a)**2)
    mask[distance_map < radius - 1 ] = 1

    mask[(distance_map > radius - 1 - dist_thr_inside_edge) & (distance_map < radius - 1)] = flag_for_dist_thr_inside_edge

    mask[(distance_map > radius - 1 ) & (distance_map < radius + 1)] = edge_flag

    # cropping the whole image
    mask[0:map_cropping, :] = map_cropping_flag
    mask[img.shape[0]-map_cropping:img.shape[0], :] = map_cropping_flag
    mask[:, 0:map_cropping] = map_cropping_flag
    mask[:, img.shape[1]-map_cropping:img.shape[1]] = map_cropping_flag

    return mask



import numpy as np
from scipy import ndimage

def canny(lp_filtered_img, canny_kernel, threshold=0.4):
    # Apply Gaussian smoothing
    smoothed = ndimage.gaussian_filter(lp_filtered_img, sigma=canny_kernel)
    
    # Calculate gradients using Sobel
    dx = ndimage.sobel(smoothed, axis=0)
    dy = ndimage.sobel(smoothed, axis=1)
    
    # Calculate gradient magnitude
    gradient_magnitude = np.hypot(dx, dy)
    
    # Normalize gradient magnitude
    gradient_magnitude = gradient_magnitude / gradient_magnitude.max()
    
    # Threshold
    lp_filtered_img_edge = (gradient_magnitude > threshold).astype(np.float32)
    
    # Clean up small objects
    lp_filtered_img_edge = ndimage.binary_opening(lp_filtered_img_edge)
    
    # Edge thinning using binary erosion with cross structure
    cross = np.array([[0,1,0],
                        [1,1,1],
                        [0,1,0]], dtype=bool)
    lp_filtered_img_edge = ndimage.binary_erosion(lp_filtered_img_edge, 
                                                structure=cross,
                                                iterations=20)
    return lp_filtered_img_edge


from .carbon_film_detector import detector_for_mrc
import tqdm
import numpy as np
import os
import pandas as pd

def read_dynamo_doc(doc_path):
    '''
    Read the mrc file paths from the Dynamo doc file.
    Args:
        doc_path: path to the Dynamo doc file
    Returns:
        mrc_index_paths: dict, key is the tomogram index, value is the mrc file path
    '''
    mrc_index_paths = {}
    with open(doc_path, 'r') as file:
        for line in file:
            parts = line.strip().split()
            if len(parts) > 1:
                mrc_path_index = int(parts[0])
                mrc_path = ' '.join(parts[1:])
                mrc_index_paths[mrc_path_index] = mrc_path
    return mrc_index_paths

def read_dynamo_tbl(tbl_path):
    '''
    Read the Dynamo tbl file.
    Args:
        tbl_path: path to the Dynamo tbl file
    Returns:
        df: DataFrame
    '''
    # Read the file and let pandas infer the dtypes
    df = pd.read_csv(tbl_path, sep=' ', header=None)
    return df

def save_dynamo_tbl(df, out_path):
    '''
    Save the Dynamo tbl file.
    Args:
        df: DataFrame
        out_path: path to the output file (tbl file)
    '''
    # Convert int columns to int (in case they were changed to float during processing)
    int_columns = [col for col in df.columns if col not in [6, 7, 8, 23, 24, 25]]
    df[int_columns] = df[int_columns].astype(int)

    # Save the DataFrame without index and header
    df.to_csv(out_path, sep=' ', header=None, index=None, float_format='%.5f')

def read_dynamo_tbl_tomogram_index(df):
    '''
    Read the tomogram indices from the Dynamo tbl file.
    Args:
        df: DataFrame
    Returns:
        tomogram_indices: list, tomogram indices
    '''
    # remove the duplicate tomogram indices
    tomogram_indices = df[19].unique()
    return tomogram_indices

def read_dynamo_tbl_particle_list(df, tomogram_index):
    '''
    Read the particle list from the Dynamo tbl file.
    Args:
        df: DataFrame
        tomogram_index: tomogram index
    Returns:
        df_slice: DataFrame, particle list
    '''
    # find the rows with the same tomogram index
    mask = df[19] == tomogram_index
    # get the row indices
    row_indices = np.where(mask)[0]
    # get the table slice
    df_slice = df.iloc[row_indices]
    
    return df_slice

def multi_mrc_processing_dynamo(doc_path,
                        tbl_path,
                        out_path,
                        low_pass,
                        detector_type,
                        kernel_radius,
                        sigma_color,
                        sigma_space,
                        canny_kernel,
                        diameter,
                        map_cropping,
                        dist_thr_inside_edge,
                        mode_threshold,
                        edge_quotient_threshold,
                        verbose):
    '''
    Process the mrc files from the Dynamo doc file and tbl file.
    Args:
        doc_path: path to the Dynamo doc file
        tbl_path: path to the Dynamo tbl file
        out_path: path to the output file (tbl file)
        low_pass: low pass filter (angstrom)
        detector_type: type of the detector
        kernel_radius: kernel radius (pixels)
        sigma_color: sigma color
        sigma_space: sigma space
        canny_kernel: canny kernel size
        diameter: diameter (angstrom)
        map_cropping: map cropping (pixels)
        dist_thr_inside_edge: distance inside edge (pixels)
        mode_threshold: mode threshold
        edge_quotient_threshold: edge quotient threshold
        verbose: whether to show the print information
    '''
    mrc_index_paths = read_dynamo_doc(doc_path)

    df = read_dynamo_tbl(tbl_path)

    # Create df_modified with the same dtypes as df
    df_modified = pd.DataFrame(columns=df.columns).astype(df.dtypes)

    tomogram_indices = read_dynamo_tbl_tomogram_index(df)

    # Add inner progress bar
    inner_pbar = tqdm.tqdm([], desc="Screening particles", 
                          position=1, dynamic_ncols=True, unit="ptcl", leave=True)

    try:
        for tomogram_index in tqdm.tqdm(tomogram_indices, desc="Processing tomograms", 
                                       position=0, dynamic_ncols=True, unit="tg", leave=True):
            df_slice = read_dynamo_tbl_particle_list(df, tomogram_index)
            mrc_path = mrc_index_paths[tomogram_index]

            if not os.path.exists(mrc_path):
                print(f"Warning: Skip tomogram {tomogram_index} {mrc_path} because not exist.")
                df_modified = pd.concat([df_modified, df_slice], ignore_index=True)
                continue

            mask = detector_for_mrc(mrc_path,
                                low_pass,
                                detector_type,
                                kernel_radius,
                                sigma_color,
                                sigma_space,
                                canny_kernel,
                                diameter,
                                map_cropping,
                                dist_thr_inside_edge,
                                mode_threshold,
                                edge_quotient_threshold,
                                show_fig=False,
                                verbose=verbose)

            if mask is False:
                df_modified = pd.concat([df_modified, df_slice], ignore_index=True)
                inner_pbar.reset(total=1)
                inner_pbar.bar_format = "{desc}: NA%|{bar:5} no carbon film detected {bar:5}|NA/NA [{elapsed}<NA, NA ptcls]"
                inner_pbar.update(0)
                if verbose:
                    print(f"Skip tomogram {tomogram_index} {mrc_path} because no carbon film detected.")
            else:
                if verbose:
                    print(f"Processing tomogram {tomogram_index} {mrc_path} with carbon film detected.")
                inner_pbar.bar_format = None
                inner_pbar.reset(total=len(df_slice))
                
                # screening the particles
                for _, row in df_slice.iterrows():
                    inner_pbar.update(1)
                    x = row[23]
                    y = row[24]
                    if mask[int(y), int(x)] == 1:
                        df_modified = pd.concat([df_modified, pd.DataFrame(row).T], ignore_index=True)

    except KeyboardInterrupt:
        inner_pbar.bar_format = None
        inner_pbar.close()
        raise
    finally:
        inner_pbar.close()

    # save the modified tbl file
    save_dynamo_tbl(df_modified, out_path)
    print(f"New tbl file saved to {out_path}.")



import cupy as cp
import matplotlib.pyplot as plt
import mrcfile
import numpy as np
from .canny import canny
from .low_pass_filter import low_pass_filter_gaussian
from .hough_transform import hough_transform_for_radius
from .bicanny import edge_detector, bilateral_filter
from .average_z import average_along_z
from ._utils import read_pixel_size, flag_masking
from .find_highest_density import find_highest_density

def detector_for_mrc(tg_path,
            low_pass,
            detector_type,
            kernel_radius,
            sigma_color,
            sigma_space,
            canny_kernel,
            diameter,
            map_cropping,
            dist_thr_inside_edge,
            mode_threshold,
            edge_quotient_threshold,
            show_fig,
            verbose):
    
    '''
    Carbon film detector for single mrc file.
    Args:
        tg_path: path to the mrc file
        low_pass: low pass filter (angstrom)
        detector_type: type of the detector. bicanny or canny.
        kernel_radius: kernel radius for bilateral filter
        sigma_color: sigma color for bilateral filter
        sigma_space: sigma space for bilateral filter
        canny_kernel: canny detector kernel size
        diameter: diameter of the carbon film (angstrom)
        map_cropping: size of the map cropping (pixels)
        dist_thr_inside_edge: distance threshold inside edge (pixels)
        mode_threshold: mode threshold. Compare the mode of the density inside and outside the arc detected.
        edge_quotient_threshold: edge quotient threshold. Compare the mean of the edge and the edge map.
        show_fig: whether to show the figure
        verbose: whether to print the information
    Returns:
        mask or False
    '''

    # read mrc file
    with mrcfile.open(tg_path) as mrc:
        data = mrc.data
    pixel_size = read_pixel_size(tg_path)
    data = cp.array(data)

    # average along z
    if data.ndim == 3:
        average = average_along_z(tg_path)
    else:
        average = data

    # low pass filter
    lp_filtered_img = low_pass_filter_gaussian(average, low_pass, pixel_size)

    if detector_type == 'bicanny':
        # bilateral filter
        lp_bi_img = bilateral_filter(lp_filtered_img, kernel_radius, sigma_color, sigma_space)
        lp_filtered_img_edge = edge_detector(lp_bi_img)
    elif detector_type == 'canny':
        # Normalize the image
        lp_filtered_img_normalized = (lp_filtered_img - lp_filtered_img.min()) / (lp_filtered_img.max() - lp_filtered_img.min())
        lp_filtered_img_edge = canny(lp_filtered_img_normalized, canny_kernel, threshold=0.4)
        lp_filtered_img_edge = lp_filtered_img_edge.astype(np.float32)
    else:
        raise ValueError(f"Detector type {detector_type} not supported!")

    # calculate the radius
    radius = int((diameter / 2) / pixel_size)

    # hough transform for the fixed radius
    center_x, center_y, hough_img = hough_transform_for_radius(lp_filtered_img_edge, radius)
    a, b = center_y-radius, center_x-radius
    
    # define the flags for the mask to mark the different regions
    map_cropping_flag = 2
    edge_flag = 3
    flag_for_dist_thr_inside_edge = 4

    # generate the mask
    mask = flag_masking(lp_filtered_img, a, b, radius, map_cropping, dist_thr_inside_edge, map_cropping_flag, edge_flag, flag_for_dist_thr_inside_edge)

    # calculate the mode difference
    mode_diff = find_highest_density(lp_filtered_img, mask.get(), inside_flags=(1, flag_for_dist_thr_inside_edge), outside_flags=(0,))

    if show_fig:
        show_figure(tg_path, hough_img, lp_filtered_img_edge, lp_filtered_img, mask, center_x, center_y, radius)
    
    if verbose:
        print(f"Carbon film edge arc: center_x {center_x}, center_y {center_y}")

    if mode_diff < mode_threshold:
        if verbose:
            print(f"mode_diff: {mode_diff} < {mode_threshold} \n Carbon film not detected!")
        return False
    else:
        if verbose:
            print(f"mode_diff: {mode_diff} > {mode_threshold} \n")

        # calculate edge_quotient
        edge_map_mean = np.mean(lp_filtered_img_edge)
        masked_edge_map_mean = np.mean(lp_filtered_img_edge[mask.get() == edge_flag])

        if verbose:
            print(f"lp_filtered_img_edge mean: {edge_map_mean}")
            print(f"masked lp_filtered_img_edge mean: {masked_edge_map_mean}")
        edge_quotient = masked_edge_map_mean / edge_map_mean

        if edge_quotient > edge_quotient_threshold:
            if verbose:
                print(f"edge_quotient: {edge_quotient} > {edge_quotient_threshold} \n Carbon film detected!")
            return mask
        else:
            if verbose:
                print(f"edge_quotient: {edge_quotient} < {edge_quotient_threshold} \n Carbon film not detected!")
            return False

        

def show_figure(tg_path, hough_img, lp_filtered_img_edge, lp_filtered_img, mask, center_x, center_y, radius):

    '''
    Show the figure for the carbon film detector.
    Args:
        hough_img: hough image
        lp_filtered_img_edge: edge map
        lp_filtered_img: low pass filtered image
        mask: mask
        center_x: center x
        center_y: center y
        radius: radius
    '''
    
    a,b = center_y-radius, center_x-radius
    theta = np.linspace(0, 2 * np.pi, num=360, dtype=np.float32)
    x = np.round(a + radius * np.cos(theta))
    y = np.round(b + radius * np.sin(theta))
    valid_indices = (x >= 0) & (x < lp_filtered_img.shape[0]) & (y >= 0) & (y < lp_filtered_img.shape[1])

    fig, axs = plt.subplots(1, 4, figsize=(20, 5), num=f'readmrc - {tg_path} - cryoCOFI')

    fig.suptitle(f'file: {tg_path}', fontsize=12)
    
    axs[0].imshow(hough_img, cmap='gray')
    axs[0].plot(center_x, center_y, 'ro')
    axs[0].set_title('Hough Space')

    axs[1].imshow(lp_filtered_img_edge, cmap='gray')
    axs[1].plot(y[valid_indices], x[valid_indices], 'r-')
    axs[1].set_title('Edge Map')

    axs[2].imshow(lp_filtered_img, cmap='gray')
    axs[2].plot(y[valid_indices], x[valid_indices], 'r-')
    axs[2].set_title('Detection Result')

    axs[3].imshow(mask.get(), cmap='gray')
    axs[3].set_title('Mask')

    plt.tight_layout()

    # Add caption for the mask image about the different flags
    captions = [
        'Flag 0: Outside Arc (excluded)',
        'Flag 1: Inside Arc (preserved)',
        'Flag 2: Map cropping (excluded)',
        'Flag 3: Edge (excluded)',
        'Flag 4: Distance inside edge (excluded)',
        '@ cryoCOFI'
    ]
    
    fig.text(0.5, 0.03, '; '.join(captions), ha='center', va='center', color='black')
    
    plt.subplots_adjust(bottom=0.1, top=0.9)
    plt.show()

if __name__ == '__main__':
    detector_for_mrc()
